function [Y, Weights, T, mu] = fastICA(X,r,type,flag)
%
% Syntax:       Y = fastICA(X,r);
%               Y = fastICA(X,r,type);
%               Y = fastICA(X,r,type,flag);
%               [Y, Weights, T, mu] = fastICA(X,r);
%               [Y, Weights, T, mu] = fastICA(X,r,type);
%               [Y, Weights, T, mu] = fastICA(X,r,type,flag);
%               
% Inputs:       X is a d x n matrix containing n samples of d-dimensional
%               data
%               
%               r is the number of independent components to compute
%               
%               [OPTIONAL] type = {'kurtosis','negentropy'} specifies
%               which flavor of non-Gaussianity to maximize. The default
%               value is type = 'kurtosis'
%               
%               [OPTIONAL] flag determines what status updates to print
%               to the command window. The choices are
%                   
%                       flag = 0: no printing
%                       flag = 1: print iteration status
%               
% Outputs:      Y is an r x n matrix containing the r independent
%               components - scaled to variance 1 - of the input samples
%               
%               Weights and T are the ICA transformation matrices such that
%               Xr = T \ Weights' * Y + repmat(mu,1,n); Note \ is a matlab
%               operation
%               is the r-dimensional ICA approximation of X

%% NOTE THAT W HERE IS NOT THE UNMIXING MATRIX               
%               
%               mu is the d x 1 sample mean of X
%               
% Description:  Performs independent component analysis (ICA) on the input
%               data using the Fast ICA algorithm
%               
% Reference:    Hyvrinen, Aapo, and Erkki Oja. "Independent component
%               analysis: algorithms and applications." Neural networks
%               13.4 (2000): 411-430
%               

% Constants
TOL = 1e-6;         % Convergence criteria
MAX_ITERS = 100;    % Max # iterations

% Parse inputs
if ~exist('flag','var') || isempty(flag)
    % Default display flag
    flag = 1;
end
if ~exist('type','var') || isempty(type)
    % Default type
    type = 'kurtosis';
end
n = size(X,2);

% Set algorithm type
if strncmpi(type,'kurtosis',1)
    % Kurtosis
    USE_KURTOSIS = true;
    algoStr = 'kurtosis';
elseif strncmpi(type,'negentropy',1)
    % Negentropy
    USE_KURTOSIS = false;
    algoStr = 'negentropy';
else
    % Unsupported type
    error('Unsupported type ''%s''',type);
end

% Center and whiten data
[Xc, mu] = centerRows(X);
[Xcw, T] = whitenRows(Xc);

% Normalize rows to unit norm
normRows = @(X) bsxfun(@rdivide,X,sqrt(sum(X.^2,2)));

% Perform Fast ICA
if flag
    % Prepare status updates
    fmt = sprintf('%%0%dd',ceil(log10(MAX_ITERS + 1)));
    str = sprintf('Iter %s: max(1 - |<w%s, w%s>|) = %%.4g\\n',fmt,fmt,fmt);
    fprintf('***** Fast ICA (%s) *****\n',algoStr);
end
Weights = normRows(rand(r,size(X,1))); % Random initial weights
k = 0;
delta = inf;
while delta > TOL && k < MAX_ITERS
    k = k + 1;
    
    % Update weights
    Wlast = Weights; % Save last weights
    Sk = Weights * Xcw;
    if USE_KURTOSIS
        % Kurtosis
        %% define g as the first derivative of the cost function in terms of Sk

        %% define gp as the second derivative of the cost function in terms of Sk

    else
        % Negentropy
        %% define g as the first derivative of the cost function in terms of Sk

        %% define gp as the second derivative of the cost function in terms of Sk

    end
    Weights = (g * Xcw') / n - bsxfun(@times,mean(gp,2),Weights);
    Weights = normRows(Weights);
    
    % Decorrelate weights
    [U, Y, ~] = svd(Weights,'econ');
    Weights = U * diag(1 ./ diag(Y)) * U' * Weights;
    
    % Update convergence criteria
    delta = max(1 - abs(dot(Weights,Wlast,2)));
    if flag
        fprintf(str,k,k,k - 1,delta);
    end
end
if flag
    fprintf('\n');
end

% Independent components
Y = Weights * Xcw;
